#ifndef __MATHLIB_H__
#define __MATHLIB_H__
#if _MSC_VER > 1000
#	pragma once
#endif	// #if _MSC_VER > 1000

#include <math.h>


#define m_PI			3.14159265358979323846f
#define m_DEG2RAD		(m_PI / 180.0f)
#define m_RAD2DEG		(180.0f / m_PI)

// evil-magic Reciprocal Square Root calculation lifted from Q3 source code
inline float Rsqrt( float x )
{
	int i;
	float y, r;

	y = x * 0.5f;
	i = *reinterpret_cast<int *>( &x );
	i = 0x5f3759df - ( i >> 1 );
	r = *reinterpret_cast<float *>( &i );
	r = r * ( 1.5f - r * r * y );

	return r;
}


template <typename T>
class _vec3
{
public:

	union {
		struct {
			T x, y, z;
		};
		T v[3];
	};

	_vec3() : x(T(0)), y(T(0)), z(T(0)) { }
	_vec3( const _vec3& v ) : x(v.x), y(v.y), z(v.z) { }
	_vec3( T _x, T _y, T _z ) : x(_x), y(_y), z(_z) { }

	inline const _vec3 operator * ( T f ) const { return _vec3(x * f,y * f,z * f); }
	inline const _vec3 operator / ( T f ) const { return _vec3(x / f,y / f,z / f); }
	inline const _vec3 operator + ( const _vec3& v ) const { return _vec3(x + v.x,y + v.y,z + v.z); }
	inline const _vec3 operator - () const { return _vec3(-x,-y,-z); }
	inline const _vec3 operator - ( const _vec3& v ) const { return _vec3(x - v.x,y - v.y,z - v.z); }

	inline _vec3& operator *= ( T f ) { return *this = *this * f; }
	inline _vec3& operator /= ( T f ) { return *this = *this / f; }
	inline _vec3& operator += ( const _vec3& v ) { return *this = *this + v; }
	inline _vec3& operator -= ( const _vec3& v ) { return *this = *this - v; }

	inline T operator * ( const _vec3& v ) const { return x * v.x + y * v.y + z * v.z; }

	inline operator T*() { return v; }
	inline operator const T*() const { return v; }

	inline T& operator[](int i) { return v[i]; }
	inline const T operator[](int i) const { return v[i]; }

	inline T length( void ) const { return (T)sqrt(x * x + y * y + z * z); }
	inline void normalize( void )
	{
		T inv = (T)Rsqrt(x * x + y * y + z * z);
		x *= inv;
		y *= inv;
		z *= inv;
	}
};

template <typename T>
inline T dot( const _vec3<T>& v1,const _vec3<T>& v2 )
{
	return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}
template <typename T>
inline _vec3<T> cross( const _vec3<T>& v1,const _vec3<T>& v2 )
{
	return _vec3<T>( v1.y * v2.z - v1.z * v2.y,
					 v1.z * v2.x - v1.x * v2.z,
					 v1.x * v2.y - v1.y * v2.x );
}

typedef _vec3<float>	vec3, vec3f;
typedef _vec3<double>	vec3d;



template <typename T>
class _mat4
{
public:

	union {
		struct {
			T	m00, m10, m20, m30,
				m01, m11, m21, m31,
				m02, m12, m22, m32,
				m03, m13, m23, m33;
		};
		T mat[16];
	};

	_mat4()
	{
		mat[0] = 1.0; mat[4] = 0.0; mat[8] = 0.0; mat[12] = 0.0;
		mat[1] = 0.0; mat[5] = 1.0; mat[9] = 0.0; mat[13] = 0.0;
		mat[2] = 0.0; mat[6] = 0.0; mat[10] = 1.0; mat[14] = 0.0;
		mat[3] = 0.0; mat[7] = 0.0; mat[11] = 0.0; mat[15] = 1.0;
	}
	_mat4( const _mat4& m )
	{
		mat[0] = m[0]; mat[4] = m[4]; mat[8] = m[8]; mat[12] = m[12];
		mat[1] = m[1]; mat[5] = m[5]; mat[9] = m[9]; mat[13] = m[13];
		mat[2] = m[2]; mat[6] = m[6]; mat[10] = m[10]; mat[14] = m[14];
		mat[3] = m[3]; mat[7] = m[7]; mat[11] = m[11]; mat[15] = m[15];
	}

	inline _mat4 operator * ( const _mat4& m ) const
	{
		mat4 ret;
		ret[0] = mat[0] * m[0] + mat[4] * m[1] + mat[8] * m[2] + mat[12] * m[3];
		ret[1] = mat[1] * m[0] + mat[5] * m[1] + mat[9] * m[2] + mat[13] * m[3];
		ret[2] = mat[2] * m[0] + mat[6] * m[1] + mat[10] * m[2] + mat[14] * m[3];
		ret[3] = mat[3] * m[0] + mat[7] * m[1] + mat[11] * m[2] + mat[15] * m[3];
		ret[4] = mat[0] * m[4] + mat[4] * m[5] + mat[8] * m[6] + mat[12] * m[7];
		ret[5] = mat[1] * m[4] + mat[5] * m[5] + mat[9] * m[6] + mat[13] * m[7];
		ret[6] = mat[2] * m[4] + mat[6] * m[5] + mat[10] * m[6] + mat[14] * m[7];
		ret[7] = mat[3] * m[4] + mat[7] * m[5] + mat[11] * m[6] + mat[15] * m[7];
		ret[8] = mat[0] * m[8] + mat[4] * m[9] + mat[8] * m[10] + mat[12] * m[11];
		ret[9] = mat[1] * m[8] + mat[5] * m[9] + mat[9] * m[10] + mat[13] * m[11];
		ret[10] = mat[2] * m[8] + mat[6] * m[9] + mat[10] * m[10] + mat[14] * m[11];
		ret[11] = mat[3] * m[8] + mat[7] * m[9] + mat[11] * m[10] + mat[15] * m[11];
		ret[12] = mat[0] * m[12] + mat[4] * m[13] + mat[8] * m[14] + mat[12] * m[15];
		ret[13] = mat[1] * m[12] + mat[5] * m[13] + mat[9] * m[14] + mat[13] * m[15];
		ret[14] = mat[2] * m[12] + mat[6] * m[13] + mat[10] * m[14] + mat[14] * m[15];
		ret[15] = mat[3] * m[12] + mat[7] * m[13] + mat[11] * m[14] + mat[15] * m[15]; 
		return ret;
	}

	inline _mat4& operator *= ( const _mat4& m ) { return *this = *this * m; }

	inline operator float*() { return mat; }
	inline operator const float*() const { return mat; }
};

typedef _mat4<float>	mat4, mat4f;
typedef _mat4<double>	mat4d;


static mat4 MakeOrtho2D( float left, float right, float bottom, float top, float zNear, float zFar )
{
	mat4 res;

	res.m00 = 2.0f / (right - left);	res.m01 = 0.0f;						res.m02 = 0.0f;						res.m03 = -(right + left) / (right - left);
	res.m10 = 0.0f;						res.m11 = 2.0f / (top - bottom);	res.m12 = 0.0f;						res.m13 = -(top + bottom) / (top - bottom);
	res.m20 = 0.0f;						res.m21 = 0.0f;						res.m22 = -2.0f / (zFar - zNear);	res.m23 = -(zFar + zNear) / (zFar - zNear);
	res.m30 = 0.0f;						res.m31 = 0.0f;						res.m32 = 0.0f;						res.m33 = 1.0f;

	return res;
}

static mat4 RotateMatrixZ( float angle )
{
	mat4 res;

	float rad = angle * m_DEG2RAD;
	float c = cosf( rad );
	float s = sinf( rad );
	res.m00 = c;	res.m01 = -s;	res.m02 = 0.0f;	res.m03 = 0.0;
	res.m10 = s;	res.m11 = c;	res.m12 = 0.0f;	res.m13 = 0.0;
	res.m20 = 0.0f;	res.m21 = 0.0f;	res.m22 = 1.0f;	res.m23 = 0.0;
	res.m30 = 0.0f;	res.m31 = 0.0f;	res.m32 = 0.0f;	res.m33 = 1.0;

	return res;
}


#endif // __MATHLIB_H__
